'''
The host part of remote Python interpreter for EventScripts for Source.
'''
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
import code
import hashlib
import select
import socket
import sys
import time

import es

try:
    bytearray
except NameError:
    # no built-in bytearray, use a bad replacement
    from .compatibility import bytearray

def load():
    '''Addon loaded using es_load.'''
    # Edit here the details for listening socket:
    family = socket.AF_INET
    address = ('127.0.0.1', 14726)
    password = str(es.ServerVar('rcon_password')) or "random password"
    # the rest is then mostly non-user-serviceable code :)
    start_listening(family, address, password)
    # HINT: it is possible to listen to multiple sockets, just call
    # start_listening several times with different parameters


def unload():
    '''Addon unloaded using es_unload.'''
    es.dbgmsg(0, 'Closing all client connections...')
    stop_serving()
    es.dbgmsg(0, 'Shutting down Python shell hosts...')
    stop_listening()
    es.dbgmsg(0, 'Goodbye, world.')


# the global data
SOCKETS = {
    'listen': set(),
    'connected': set(),
    'active': False,
}

info = es.AddonInfo()
info.basename = 'remote_interpreter'
info.name = 'Remote Python Interpreter host for Source servers'
info.author = 'GODJonez'
info.version = "1.4"


def start_listening(family, address, password):
    '''
    Create a socket and start listening for connections on it, requiring
    specified password.
    '''
    lsocket = socket.socket(family, socket.SOCK_STREAM)
    lsocket.bind(address)
    lsocket.listen(0) # only expected one new connection at a time
    pwhash = hashlib.sha1(password).hexdigest()
    if not SOCKETS['listen']:
        _start_checking()
    SOCKETS['listen'].add(ListeningSocket(lsocket, pwhash))


def _start_checking():
    '''
    Start tick listener that checks network sockets.
    '''
    if not SOCKETS['active']:
        es.addons.registerTickListener(check_sockets)


def _stop_checking():
    '''
    Start tick listener that checks network sockets.
    '''
    if SOCKETS['active']:
        es.addons.unregisterTickListener(check_sockets)


def stop_listening():
    '''
    Stops all listening sockets from accepting new connections.
    '''
    for lsocket in list(SOCKETS['listen']):
        lsocket.close()
    if not SOCKETS['connected']:
        _stop_checking()


def stop_serving():
    '''
    Closes all connected sockets.
    '''
    for csocket in list(SOCKETS['connected']):
        csocket.disconnect('Stopping serving clients.')
    if not SOCKETS['listen']:
        _stop_checking()


def check_sockets():
    '''
    Called by EventScripts on each gameframe whenever there are sockets here
    waiting for something to happen.
    '''
    want_read = list(SOCKETS['listen'] | SOCKETS['connected'])
    if not want_read:
        # no want read :(
        # at least stop calling me then!
        _stop_checking()
        return
    want_write = filter(None, SOCKETS['connected'])
    will_read, will_write, will_error = select.select(
        want_read, want_write, [], 0)
    for kneesock in will_read:
        try:
            kneesock.ready_to_read()
        except socket.error, error:
            es.dbgmsg(0, 'Closing a socket because of a socket error (read).')
            sys.excepthook(*sys.exc_info())
            sys.exc_clear()
            kneesock.close()
        except BaseException, error:
            sys.excepthook(*sys.exc_info())
            sys.exc_clear()
    for kneesock in will_write:
        try:
            sent = kneesock.socket.send(str(kneesock.sendbuffer))
        except socket.error, error:
            es.dbgmsg(0, 'Closing a socket because of a socket error (write).')
            sys.excepthook(*sys.exc_info())
            sys.exc_clear()
            kneesock.close()
        else:
            kneesock.sendbuffer[:sent] = []


class ListeningSocket(object):
    '''
    A little convenience wrapper to help managing multiple listening points.
    '''
    def __init__(self, lsocket, pwhash):
        self.socket = lsocket
        self.pwhash = pwhash
    
    def fileno(self):
        '''
        Make this appear as if it was real socket by giving fileno of the
        associated socket, allowing this object to be used in select call.
        '''
        return self.socket.fileno()
    
    def close(self):
        '''
        Close the socket.
        '''
        es.dbgmsg(0, 'Listening closed.')
        self.socket.close()
        SOCKETS['listen'].remove(self)
    
    def ready_to_read(self):
        '''
        Called by check_sockets indicating someone is connecting here.
        '''
        csocket, caddress = self.socket.accept()
        es.dbgmsg(0, '[Remote Python Interpreter] New connection: %r' %
            (caddress,))
        client = NetworkInterpreter(csocket, caddress, self.pwhash)
        SOCKETS['connected'].add(client)


class NetworkInterpreter(code.InteractiveInterpreter):
    '''
    A Python interpreter working across network using a custom protocol.
    '''
    def __init__(self, csocket, caddress, pwhash):
        self._original_outputs = None
        self.socket = csocket
        self.address = caddress
        self.pwhash = pwhash
        self.salt = hashlib.sha1(str(time.time())).hexdigest()
        self.sendbuffer = bytearray(self.salt+'\n')
        self.recvbuffer = bytearray()
        self._input = bytearray()
        # FIXME NOTE: How to wrap stdout and stderr of this subinterp?
        # They will likely go to local console and not to client...
        code.InteractiveInterpreter.__init__(self)
    
    def close(self):
        '''
        Close the socket.
        '''
        es.dbgmsg(0, '[Remote Python Interpreter] Connection closed.')
        self.socket.close()
        SOCKETS['connected'].remove(self)
    
    def disconnect(self, message):
        '''
        Close the socket after explaining why.
        '''
        try:
            es.dbgmsg(0, '[Remote Python Interpreter] Disconnecting: '+message)
            if not self.salt:
                self.socket.send(self._make_message('C', message))
        finally:
            # whatever happens there, close the damn socket
            self.close()
    
    def fileno(self):
        '''
        Make this appear as if it was real socket by giving fileno of the
        associated socket, allowing this object to be used in select call.
        '''
        return self.socket.fileno()
    
    def ready_to_read(self):
        '''
        Called by check_sockets indicating there is new data incoming.
        '''
        es.dbgmsg(2, '[Remote Python Interpreter] Receiving data')
        incoming = self.socket.recv(4096)
        es.dbgmsg(3, repr(incoming[:200]))
        if not incoming:
            es.dbgmsg(0, '[Remote Python Interpreter] Remote closed')
            # connection closed
            self.close()
            return
        self.recvbuffer.extend(incoming)
        while '\n' in self.recvbuffer:
            boundary = self.recvbuffer.find('\n')
            line = str(self.recvbuffer[:boundary])
            es.dbgmsg(3, 'Extracted line: '+repr(line))
            self.recvbuffer[:boundary+1] = []
            self._handle_line(line)
    
    def __nonzero__(self):
        '''
        Boolean check indicating willingess to send data.
        '''
        return bool(self.sendbuffer)
    
    @staticmethod
    def _make_message(msgtype, msg):
        '''
        Make a string to be inserted into send buffer.
        '''
        if msg[-1:] == '\n':
            return '%s:%s' % (msgtype, msg)
        else:
            return '%s:%s\n' % (msgtype, msg)
    
    def _send_message(self, msgtype, msg):
        '''
        Make and send a message.
        '''
        message = self._make_message(msgtype, msg)
        es.dbgmsg(2, '[Remote Python Interpreter] Sending message ' + message)
        self.sendbuffer.extend(message)
    
    def _capture_output(self):
        '''
        Start capturing stdout and stderr to send those to network.
        '''
        self._original_outputs = sys.stdout, sys.stderr
        sys.stdout = StringIO.StringIO()
        sys.stderr = StringIO.StringIO()
    
    def _release_output(self):
        '''
        Stop capturing outputs and send any captures to network.
        '''
        out, err = sys.stdout.getvalue(), sys.stderr.getvalue()
        sys.stdout, sys.stderr = self._original_outputs
        self._original_outputs = None # just in case
        if out:
            self.write(out, 'O')
        if err:
            self.write(err, 'E')
        
    def _handle_line(self, line):
        '''
        Handle input from client.
        '''
        if self.salt:
            # expect password
            correct = hashlib.sha1(self.pwhash + self.salt).hexdigest()
            self.salt = None
            if correct == line:
                # password accepted
                self._send_message('N', 'Password OK.')
                self._send_message('1', '') # default prompt
            else:
                # invalid password
                self.disconnect('Invalid password.')
        else:
            # expect input to shell :)
            self._input.extend(line)
            self._capture_output()
            # wrap in try just to ensure we will release output
            try:
                result = self.runsource(str(self._input), '<network>')
            finally:
                self._release_output()
            if result:
                # incomplete input
                self._input.extend('\n') # new line coming up :)
                self._send_message('2', '') # continue prompt
            else:
                # complete input or error
                self._input[:] = [] # ate it all, thanks for the food
                self._send_message('1', '') # default prompt

    def write(self, data, output='E'):
        '''
        Overrides InteractiveInterpreter.write() to send output to network.

        Normally this is called with errors, so default output is E.
        '''
        lines = data.split('\n')
        if len(lines) > 1 and not lines[-1]:
            lines.pop()
        for line in lines:
            self._send_message(output, line)

